from typing import Any, Optional

from stere import Stere


class ElementStrategy(Stere):
    """Base class for a Strategy.

    Elements are used by Fields as the source of what finds the DOM content.

    Arguments:
        strategy (str): The type of strategy to use when locating an element.
        locator (str): The locator for the strategy.
        parent_locator: A parent object to search from. If None,
            search will occur from top of the DOM.
    """

    def __init__(
        self, strategy: str,
        locator: str,
        parent_locator: Optional[Any] = None,
    ):
        self.strategy = strategy
        self.locator = locator
        self.parent_locator = parent_locator

        # A Field that should be searched for and whose element should be set
        # as the parent, but only when .find() is called.
        self.root: Optional[Any] = None

    def _find_all(self, wait_time: Optional[int] = None):
        """Find from inside a parent element."""
        raise NotImplementedError

    def find(self, wait_time: Optional[int] = None):
        """Use ElementStrategy._find_all() to find an element.

        If a root has been set, it will set the parent_locator attribute
        before searching.
        """
        if self.root:
            self.parent_locator = self.root.find(wait_time=wait_time)
        return self._find_all(wait_time=wait_time)



import copy

from .area import Area
from .areas import Areas
from .repeating import Repeating
from ..fields import Field


class RepeatingArea(Repeating):
    """
    Represents multiple identical Areas on a page.

    A root argument is required, which is expected to be a non-unique Field
    on the page.

    A collection of Areas are built from every instance of the root that is
    found. Every other Field provided in the arguments is populated inside
    each Area.

    In the following example, there's a table with 15 rows. Each row has
    two cells. The sixth row in the table should have an item with the
    name "Banana" and a price of "$7.00"

    >>> from stere.areas import RepeatingArea
    >>> from stere.fields import Root, Link, Text
    >>>
    >>> class Inventory(Page):
    >>>     def __init__(self):
    >>>         self.inventory_items = RepeatingArea(
    >>>             root=Root('xpath', '//table/tr'),
    >>>             name=Link('xpath', './td[1]'),
    >>>             price=Text('xpath', './td[2]'),
    >>>         )

    >>> inventory = Inventory()
    >>> assert 15 == len(inventory.areas)
    >>> assert "Banana" == inventory.areas[5].name
    >>> assert "$7.00" == inventory.areas[5].price
    """

    def __init__(self, root: Field, **kwargs):
        self.root = root
        self.repeater = Area
        self.repeater_name = self.repeater.__name__

        self._items = {}
        for k, v in kwargs.items():
            if not isinstance(v, (Field, Area)):
                raise ValueError(
                    'RepeatingArea arguments can only be a Field or Area.',
                )
            if k != 'root':
                self._items[k] = v
                # Field (in plural) can be accessed directly.
                setattr(self, f'{k}s', v)

    def new_container(self) -> Areas:
        """Get a new instance of the container this class uses.

        Returns:
            Areas

        """
        return Areas()

    @property
    def areas(self) -> Areas:
        """Find all instances of the root,
        then return a list of Areas: one for each root.

        Returns:
            Areas: list-like collection of every Area that was found.

        Example:

            >>> def test_stuff():
            >>>     listings = MyPage().my_repeating_area.areas
            >>>     listings[0].my_input.fill('Hello world')

        """
        return self.children()

    def children(self) -> Areas:
        """Find all instances of the root,
        then return a list of Areas: one for each root.

        Returns:
            Areas: list-like collection of every Area that was found.

        Example:

            >>> def test_stuff():
            >>>     listings = MyPage().my_repeating_area.areas
            >>>     listings[0].my_input.fill('Hello world')

        """
        all_roots = self._all_roots()
        container = self.new_container()

        for root in all_roots:
            copy_items = copy.deepcopy(self._items)
            for field_name in copy_items.keys():
                child = copy_items[field_name]

                child._set_parent_locator(root)

            new_area = self.repeater(**copy_items)
            container.append(new_area)
        return container



# Javascript intended to be opened as a string and injected into the browser.
js_script = """
// Prevent patch from stacking if applied multiple times.
if (!window.oldSend) {
    // Number of requests that are active.
    document.activeXHRrequests = 0;

    // Number of requests that have been completed.
    document.totalXHRrequests = 0;

    window.oldSend = XMLHttpRequest.prototype.send;

    // Patch a request counter onto XMLHttpRequest.send()
    XMLHttpRequest.prototype.send = function() {
        countRequests(this);
        // Call the native send()
        window.oldSend.apply(this, arguments);
    }

    // Increment a counter every time a request starts,
    //   and add an event to decrement it when the request is complete.
    var countRequests = function(xhr) {
        document.activeXHRrequests += 1;
        document.totalXHRrequests += 1;

        xhr.addEventListener(
            'loadend', () => {
                // Decrement counter
                document.activeXHRrequests -= 1;

                // For live debugging
                var remaining_msg = `remaining: ${document.activeXHRrequests}`;
                var total_msg = `total: ${document.totalXHRrequests}`;
                console.log('XHR loadend,', remaining_msg, total_msg);
            }
        )
    }
}
"""



from .repeating import Repeating
from ..fields import Field


class Area:
    """A collection of unique fields.

    The Area object takes any number of Fields as arguments.
    Each Field must be unique on the Page and only present in one Area.

    Example:

    >>> from stere.areas import Area
    >>> from stere.fields import Button
    >>>
    >>> class Album(Page):
    >>>     def __init__(self):
    >>>         self.tracks = Area(
    >>>             first_track=Button('xpath', '//my_xpath_string'),
    >>>             second_track=Button('xpath', '//my_xpath_string'),
    >>>             third_track=Button('xpath', '//my_xpath_string'),
    >>>         )
    >>>
    >>> def test_stuff():
    >>>     album = Album()
    >>>     album.tracks.third_track.click()
    """

    def __init__(self, root=None, **kwargs):
        self.root = root

        # Store kwargs
        self._items = {}

        for key, value in kwargs.items():
            if not isinstance(value, (Field, Area, Repeating)):
                raise ValueError(
                    (
                        'Areas must only be initialized with: '
                        'Field, Area, Repeating types'
                    ),
                )
            self._items[key] = value

            # Sets the root for the element, if provided.
            if self.root is not None and value is not self.root:
                # Set the Field's _element's root
                if isinstance(value, Field):
                    value._element.root = self.root

                elif isinstance(value, Area):
                    # Area has a root, give that root a root.
                    if value.root is not None:
                        value.root._element.root = self.root
                    # Area has no root, set every Field's _element's root
                    else:
                        for _k, v in value._items.items():
                            v._element.root = self.root

                # Repeating sets its root Field's root.
                elif isinstance(value, Repeating):
                    value.root._element.root = self.root

            # Field can be called directly.
            setattr(self, key, value)

        self._workflow = None

    def _set_parent_locator(self, element):
        """For every item in the Area, set a parent_locator."""
        if self.root is not None:
            self.root._set_parent_locator(element)
        else:
            for _, v in self._items.items():
                v._set_parent_locator(element)

    def workflow(self, value: str):
        """Set the current workflow for an Area.

        Designed for chaining before a call to perform().

        Arguments:
            value (str): The name of the workflow to set.

        Returns:
            Area: The calling Area

        Example:

        >>> my_area.workflow('Foobar').perform()

        """
        self._workflow = value
        return self

    def perform(self, *args, **kwargs):
        """For every Field in an Area, "do the right thing"
        by calling the Field's perform() method.

        Fields that require an argument can either be given sequentially
        or with keywords.

        Arguments:
            args: Arguments that will sequentially be sent to Fields
                in this Area.
            kwargs: Arguments that will be sent specifically to the Field
                with a matching name.

        Example:

            Given the following Page Object:

            >>> from stere.areas import Area
            >>> from stere.fields import Button, Input
            >>>
            >>> class Login():
            >>>     def __init__(self):
            >>>         self.form = Area(
            >>>             username=Input('id', 'app-user'),
            >>>             password=Input('id', 'app-pwd'),
            >>>             submit=Button('id', 'app-submit')
            >>>         )

            Any of the following styles are valid:

            >>> def test_login():
            >>>     login = Login()
            >>>     login.my_area.perform('Sven', 'Hoek')

            >>> def test_login():
            >>>     login = Login()
            >>>     login.my_area.perform(username='Sven', password='Hoek')

            >>> def test_login():
            >>>     login = Login()
            >>>     login.my_area.perform('Sven', password='Hoek')
        """
        arg_index = 0
        workflow = self._workflow
        for field_name, field in self._items.items():
            # If the Field isn't in the current workflow, skip it entirely.
            if workflow is not None and workflow not in field.workflows:
                continue

            if field_name in kwargs:
                result = field.perform(kwargs[field_name])
            else:
                if args:
                    result = field.perform(args[arg_index])
                else:
                    result = field.perform()

                # If we've run out of arguments, don't increase the index.
                if field.consumes_arg and len(args) > (arg_index + 1):
                    arg_index += 1

        self._workflow = None

        return result



import copy
from typing import Any, Optional

from ..utils import _retry


class Repeating:
    """
    Arguments:
        root (Field): A non-unique root to search for.
        repeater (Repeating): An object that inherits from Repeating.

    Represents abstract non-unique collections that repeat,
    based on a common root.

    This can be used to identify anything that not only appears multiple times,
    but also contains things which appear multiple times.

    Repeating are inherently confusing and should only be used if something
    appears multiple times, contains something else that appears multiple
    times, and is truly non-unique with no other way to target it.

    The last object in a chain of Repeating must be a RepeatingArea.
    This is because ultimately, there must be an end to the number of things
    that repeat.

    Example:

        A page where multiple "project" containers appear, each with a
        table of items.

        >>> from stere.areas import Repeating, RepeatingArea
        >>> from stere.fields import Root, Link, Text
        >>>
        >>> projects = Repeating(
        >>>     root=Root('css', '.projectContainer'),
        >>>     repeater=RepeatingArea(
        >>>         root=Root('xpath', '//table/tr'),
        >>>         description=Link('xpath', './td[1]'),
        >>>         cost=Text('xpath', './td[2]'),
        >>>    )
        >>> )
        >>>
        >>> assert 2 == len(projects.children)
        >>> first_project = projects.children[0]
        >>> assert first_project.areas.contains(
        >>>     'description', 'Solar Panels')
        >>>
        >>> second_project = projects.children[1]
        >>> assert second_project.areas.contains(
        >>>     'description', 'Self-Driving Cars')

    """

    def __init__(self, root, repeater):
        self.root = root
        self.repeater = repeater
        self.repeater_name = type(self.repeater).__name__

    def _set_parent_locator(self, element):
        """Set the parent_locator of the root."""
        self.root._element.parent_locator = element

    def new_container(self) -> Any:
        """Must return an object to contain results from Repeater.children()

        By default a list is returned.

        Returns:
            list

        """
        return []

    def __len__(self) -> int:
        """Return the number of times the root was found.

        Does not actually build the children.

        """
        all_roots = self.root.find_all()
        return len(all_roots)

    def _all_roots(self):
        """Search for all instances of the root.

        Only intended for use inside Repeating.children().
        Otherwise use self.root.find_all().

        Raises:
            ValueError: If no instances of the root were found.

        Returns:
            All instances of the Repeating's root

        """
        all_roots = self.root.find_all()
        if 0 == len(all_roots):
            raise ValueError(
                f'Could not find any {self.repeater_name} using the root: '
                f'{self.root.locator}',
            )
        return all_roots

    def children(self) -> Any:
        """Find all instances of the root,
        then return a collection containing children built from those roots.

        The type of collection is determined by the Repeating.new_container()
        method.

        Returns:
            list-like collection of every repeater that was found.

        """
        all_roots = self._all_roots()
        container = self.new_container()

        for root in all_roots:
            copy_repeater = copy.deepcopy(self.repeater)
            # Set the repeater's parent locator to the found root instance
            copy_repeater._set_parent_locator(root)

            container.append(copy_repeater)

        return container

    def has_children(
        self,
        minimum: int = 1,
        retry_time: Optional[int] = None,
    ) -> bool:
        """Check if any children can be found.

        Arguments:
            minimum: Minimum number of children that must exist.
            retry_time: Number of seconds to check for.

        Returns:
            bool

        """
        return _retry(
            lambda: len(self) >= minimum,
            retry_time=retry_time,
        )



import time
import typing
from functools import reduce

from stere import Stere


def rgetattr(obj, attr, *args):
    """A nested getattr"""
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return reduce(_getattr, [obj] + attr.split('.'))


def _retry(
    fn: typing.Callable, retry_time: typing.Optional[int] = None,
) -> bool:
    """Retry a function for a specific amount of time.

    Returns:
        True if the function returns a truthy value, else False

    Arguments:
        fn (function): Function to retry
        retry_time: Number of seconds to retry. If not specified,
            Stere.retry_time will be used.

    """
    retry_time = retry_time or Stere.retry_time
    end_time = time.time() + retry_time

    while time.time() < end_time:
        if fn():
            return True
    return False



import configparser
import os
import typing


def _parse_config() -> configparser.ConfigParser:
    parser = configparser.ConfigParser()
    cwd = os.getcwd()
    file_path = f'{cwd}/stere.ini'
    parser.read(file_path)
    return parser


def _get_config_option(
    parser: configparser.ConfigParser,
    section: str = 'stere',
    option: str = '',
    default: typing.Optional[typing.Any] = '',
) -> typing.Any:
    """Get an option from a config section, if it exists.

    Arguments:
        section(str): The name of the section
        option(str): The name of the option

    Returns:
        str: The found value, or else the default value

    """
    if parser.has_option(section, option):
        return parser.get(section, option)
    return default


class BrowserEnabled:
    """Base class that stores attributes at the class level,
    shared by every object that inherits from this class.

    If a stere.ini file is found, the attributes will be set from there.

    Attributes:
        browser (object): Pointer to what is driving the automation.
        base_url (str): Used as the url when navigating to pages.
        url_suffix (str): Appended to base_url when navigating to pages.
        retry_time (int): The maximum amount of time in seconds to try and
            find an element on a page.
        library (str): Name of the automation library to use. Default is
            splinter.
        url_navigator (str): Name of the function that opens a page.

    """

    browser = None
    base_url: str = ''
    url_suffix: str = ''
    retry_time: int = 5

    xhr_spy_sleep_time = 1
    fetch_spy_sleep_time = 1

    # Default values for automation libraries
    library_defaults = {
        'splinter': {
            'url_navigator': 'visit',
        },
    }

    library: str = 'splinter'
    url_navigator: str = library_defaults[library]['url_navigator']

    # If a config file exists, get settings from there.
    parser = _parse_config()
    if parser.has_section('stere'):
        library = _get_config_option(parser, option='library', default=library)
        url_navigator = _get_config_option(
            parser, option='url_navigator', default=url_navigator,
        )
        base_url = _get_config_option(
            parser, option='base_url', default=base_url,
        )
        url_suffix = _get_config_option(
            parser, option='url_suffix', default=url_suffix,
        )
        retry_time = _get_config_option(
            parser, option='retry_time', default=retry_time,
        )



import urllib.parse

from .browser_spy import FetchSpy, XHRSpy
from .browserenabled import BrowserEnabled


class Page(BrowserEnabled):
    """Represents a single page in an application.
    The Page class is the base which all Page Objects should inherit from.

    Inheriting from Page is not required for Fields or Areas to work.

    All attribute calls that fail are then tried on the browser attribute.
    This allows classes inheriting from Page to act as a proxy to
    whichever browser/driver is being used.

    Using Splinter's browser.url method as an example, the following methods
    are analogous:

    >>> MyPage.url == MyPage.browser.url == browser.url

    The choice of which syntax to use depends on how you want to write your
    test suite.
    """

    # Allows network requests to be spied on
    fetch_spy = FetchSpy()
    xhr_spy = XHRSpy()

    def __getattr__(self, val):
        """If an attribute doesn't exist, try getting it from the browser."""
        return getattr(self.browser, val)

    def __enter__(self):
        """Page Objects can be used as context managers."""
        return self

    def __exit__(self, *args):
        """Page Objects can be used as context managers."""
        pass

    @property
    def page_url(self) -> str:
        """Get a full URL from stere's base_url and a Page's url_suffix.

        Uses urllib.parse.urljoin to combine the two.
        """
        return urllib.parse.urljoin(self.base_url, self.url_suffix)

    def navigate(self):
        """When the base Stere object has been given the `url_navigator`
        attribute, and a Page Object has a `page_url` attribute, the
        `navigate()` method can be called.

        This method will call the method defined in `url_navigator`,
        with `page_url` as the first parameter.

        Returns:
            Page: The instance where navigate() was called from.

        Example:

            >>> from splinter import Browser
            >>> from stere import Page
            >>>
            >>>
            >>> class Home(Page):
            >>>     def __init__(self):
            >>>         self.page_url = 'https://en.wikipedia.org/'
            >>>
            >>>
            >>> Stere.browser = Browser()
            >>> Stere.url_navigator = 'visit'
            >>>
            >>> home_page = Home()
            >>> home_page.navigate()

        """
        getattr(self.browser, self.url_navigator)(self.page_url)
        return self



import copy
import typing

from selenium.common.exceptions import StaleElementReferenceException

from .element_strategy import ElementStrategy
from .strategy import strategy
from ..utils import _retry


class SplinterBase(ElementStrategy):
    def is_clickable(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is present in the DOM and clickable.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        return _retry(
            lambda: self.find() and self.find()._element.is_enabled(),
            wait_time,
        )

    def is_not_clickable(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is not clickable in the DOM.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        def search():
            result = self.find(wait_time=0)
            if not result:
                return True
            if result and not result._element.is_enabled():
                return True
            return False

        return _retry(search, wait_time)

    def is_present(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is present in the DOM.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        return _retry(
            lambda: self.find(),
            wait_time,
        )

    def is_not_present(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is not present in the DOM.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        return _retry(
            lambda: not self.find(wait_time=0),
            wait_time,
        )

    def is_visible(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is present in the DOM and visible.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        def search():
            elem = self.find()
            if elem:
                try:
                    result = elem.visible
                # StaleElementReferenceException occurs if element is found
                # but changes before visible is checked
                except StaleElementReferenceException:
                    return False

                if result:
                    return True

            return False

        return _retry(search, wait_time)

    def is_not_visible(self, wait_time: typing.Optional[int] = None) -> bool:
        """Check if an element is not visible in the DOM.

        Arguments:
            wait_time (int): The number of seconds to wait. If not specified,
                Stere.retry_time will be used.
        """
        def search():
            elem = self.find(wait_time=0)
            if elem:
                try:
                    result = elem.visible
                # StaleElementReferenceException occurs if element is found
                # but changes before visible is checked
                except StaleElementReferenceException:
                    return False

                if not result:
                    return True
            else:
                return True

            return False

        return _retry(search, wait_time)

    def _find_all(self, wait_time: typing.Optional[int] = None):
        """Find from inside a parent element."""
        parent = self.parent_locator or self.browser
        func = getattr(parent, f'find_by_{self.strategy}')
        return func(self.locator, wait_time=wait_time)


@strategy('css')
class FindByCss(SplinterBase):
    strategy = 'css'


@strategy('xpath')
class FindByXPath(SplinterBase):
    strategy = 'xpath'


@strategy('tag')
class FindByTag(SplinterBase):
    strategy = 'tag'


@strategy('name')
class FindByName(SplinterBase):
    strategy = 'name'


@strategy('text')
class FindByText(SplinterBase):
    strategy = 'text'


@strategy('id')
class FindById(SplinterBase):
    strategy = 'id'


@strategy('value')
class FindByValue(SplinterBase):
    strategy = 'value'


class FindByAttribute(SplinterBase):
    """Strategy to find an element by an arbitrary attribute."""

    _attribute = ''

    def _find_all(self, wait_time: typing.Optional[int] = None):
        """Find from inside parent element."""
        parent = self.parent_locator or self.browser
        return parent.find_by_css(
            f'[{self._attribute}="{self.locator}"]', wait_time=wait_time,
        )


def add_data_star_strategy(data_star_attribute: str):
    """Add a new splinter strategy that finds by data_star_attribute.

    Arguments:
        data_star_attribute (str): The data-* attribute to use in the new
            strategy.
    """
    find_by_data_star = copy.deepcopy(FindByAttribute)
    find_by_data_star._attribute = data_star_attribute
    return strategy(data_star_attribute)(find_by_data_star)



from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('send_keys', consumes_arg=True)
class Input(Field):
    """Convenience Class on top of Field.

    Uses Appium's send_keys method.

    Arguments:
        default_value (str): When Input.send_keys() is called with no
            arguments, this value will be used instead.
    """

    def __init__(self, *args, default_value=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.default_value = default_value

    @use_after
    @use_before
    def send_keys(self, value=None):
        """Use Appium's fill method.

        Arguments:
            value (str): The text to enter into the input.

        Example:

        >>> first_name = Input('id', 'fillme')
        >>> first_name.send_keys('Joseph')

        """
        if value is None and self.default_value:
            value = self.default_value
        self.find().send_keys(value)



strategies = {}


def strategy(strategy_name: str):
    """Register a strategy name and strategy Class.

    Use as a decorator.

    Example:
        @strategy('id')
        class FindById:
            ...

    Strategy Classes are used to build Elements Objects.

    Arguments:
        strategy_name (str): Name of the strategy to be registered.
    """
    def wrapper(finder_class):
        global strategies
        strategies[strategy_name] = finder_class
        return finder_class
    return wrapper



from typing import Callable, Dict, List


class EventEmitter:
    """A basic Event Emitter.

    Attributes:
        _events_data: A dict of every registered event and the listeners
            attached to them.
    """

    def __init__(self):
        self._events_data: Dict[str, List[Dict]] = {}

    @property
    def events(self) -> List[str]:
        """Get the names of every registered event."""
        return list(self._events_data.keys())

    def on(self, event: str, listener: Callable) -> List[Dict]:
        """Register a new listener to an event.

        If the event does not exist, it will be created.

        Arguments:
            event (str): The name of the event to register a listener to.
            listener (Callable): The function to run when the event is emitted.

        Returns:
            list: The list of listeners for the event.

        Example:

            >>> def my_event_function(emitter):
            >>>    pass
            >>>
            >>> my_field = Field()
            >>> my_field.on('before', my_event_function)

        """
        if event not in self._events_data:
            self._events_data[event] = []

        listener_data = {'listener': listener}
        self._events_data[event].append(listener_data)

        return self._events_data[event]

    def emit(self, event: str):
        """Emit an event.

        Every listener registered to the event will be called with the
        emitter class as the first argument.

        Listeners are called in the order of registration.

        Arguments:
            event (str): The name of the event to emit.
        """
        event_data = self._events_data.get(event)
        if event_data is None:
            raise ValueError(f'{event} is not a registered Event.')

        for ev in event_data:
            ev['listener'](self)



import time
import typing

from stere import Stere

from .button import Button
from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('select', consumes_arg=True)
class Dropdown(Field):
    """Represents a dropdown menu.
    If the "option" argument is provided with a field,
        use that as the dropdown item.
    Else, assume a standard HTML Dropdown and use the option tag.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # If no option arg is given, assume Dropdown is a standard HTML one.
        if kwargs.get('option') is None:
            self.option = Button('tag', 'option')
        else:
            self.option = kwargs.get('option')

    def __getitem__(self, index):
        """Accessing by index will select the option matching the index."""
        return self.options[index].click()

    @property
    def options(self):
        """Search for all the elements that are an option in the dropdown.

        Returns:
            list

        """
        self.option._element.parent_locator = self.find()
        return list(self.option.find_all())

    @use_after
    @use_before
    def select(self, value: str, retry_time: typing.Optional[int] = None):
        """Search for an option by its html content, then clicks the one
        that matches.

        Arguments:
            value (str): The option value to select.
            retry_time (int): The amount of time to try to find the value.
                Default is Stere.retry_time.

        Raises:
            ValueError: The provided value could not be found in the dropdown.

        """
        retry_time = retry_time or Stere.retry_time
        end_time = time.time() + retry_time

        found_options = []

        while time.time() < end_time:
            found_options = self._select(value)
            if not len(found_options):
                return

        raise ValueError(
            f'{value} was not found. Found values are: {found_options}')

    def _select(self, value: str) -> typing.List[str]:
        found_options = []

        for option in self.options:
            found_options.append(option.html)
            if option.html == value:
                option.click()
                return []

        return found_options



import re

from moneyed import Money as PyMoney

from ..decorators import stere_performer
from ..field import Field


@stere_performer('null_action', consumes_arg=False)
class Money(Field):
    """A simple wrapper over Field, it does not implement a performer method.

    Money has methods for handling Fields where the text is a form of currency.

    Example:

        >>> from stere.fields import Money
        >>>
        >>>
        >>> self.price = Money('id', 'item_price')
    """

    number_regex = r'[^0-9\.]+'

    def money(self, currency: str = 'USD') -> PyMoney:
        """Create a Money object from the Field's text.

        The returned object is an instance of moneyed.Money.
        See: `py-moneyed <https://github.com/limist/py-moneyed>`_

        Arguments:
            currency (str): Name of the currency to use

        Returns:
            moneyed.Money

        """
        return PyMoney(amount=self.number, currency=currency)

    @property
    def number(self) -> str:
        """The Field's text, normalized to look like a number."""
        m = re.compile(self.number_regex, re.IGNORECASE)
        return m.sub('', self.text)



import typing

import splinter

from stere import Stere

from .build_element import build_element
from .decorators import stere_performer
from ..event_emitter import EventEmitter
from ..utils import _retry


@stere_performer('null_action', consumes_arg=False)
class Field(EventEmitter):
    """The representation of individual page components.

    Conceptually, they're modelled after general behaviours, not specific
    HTML tags.

    Arguments:
        strategy (str): The type of strategy to use when locating an element.
        locator (str): The locator for the strategy.
        workflows (list): Any workflows the Field should be included with.

    Example:

        >>> from stere.fields import Field
        >>> my_field = Field('xpath', '//*[@id="js-link-box-pt"]/small/span')

    Attributes:
        element: The result of a search operation on the page.
            All attribute calls on the Field that fail are then tried on the
            element.

            This allows classes inheriting from Field to act as a proxy to the
            underlying automation library.

            Using Splinter's `visible` attribute as an example, the following
            methods are analogous:

            >>> Field.visible == Field.find().visible == Field.element.visible

    """

    def __init__(self, strategy: str, locator: str, *args, **kwargs):
        super().__init__()

        # Ensure before and after events are valid.
        self._events_data['before'] = []
        self._events_data['after'] = []

        self.strategy = strategy
        self.locator = locator
        self._element = build_element(strategy, locator)

        self.workflows = kwargs.get('workflows') or []
        self.returns = kwargs.get('returns') or None

    def __call__(self, *args, **kwargs):
        """When a Field instance is called, run the perform() method."""
        return self.perform(*args, **kwargs)

    def __getattr__(self, val):
        """If an attribute doesn't exist, try getting it from the element.

        If it still doesn't exist, do a find() on the element and see if the
        attribute exists there.
        """
        element = super().__getattribute__('_element')
        try:
            return getattr(element, val)
        except AttributeError:
            # Allows deepcopy not to get into infinite recursion.
            if val in ['__deepcopy__', '__getstate__']:
                raise AttributeError

            # Try getting the attribute from the found element.
            try:
                elem = self.find(Stere.retry_time)
            except splinter.exceptions.ElementDoesNotExist as e:
                msg = (
                    'Failed to get element attribute.'
                    f'Could not find element with {self.strategy}: {self.locator}'  # NOQA E501
                )
                raise AttributeError(msg) from e

            return getattr(elem, val)

    def __repr__(self):
        """Provide a string representation of this class."""
        return (
            f'{self.__class__.__name__} - '
            f'Strategy: {self.strategy}, Locator: {self.locator}')

    def _set_parent_locator(self, element):
        """Set the parent locator of this Field's element."""
        self._element.parent_locator = element

    @property
    def element(self):
        """Tries to find the element, then returns the results."""
        return self._element.find()

    def null_action(self) -> None:
        """Empty method used as the performer for Field.

        Allows the base Field object to be used in an Area.
        """
        pass

    def before(self) -> None:
        """Called automatically before methods with the `@use_before`
        decorator are called.

        Performer methods are decorated with @use_before.

        By default it does nothing. Override this method if an action must be
        taken before a method is called.

        In the following example, Dropdown has been subclassed to hover over
        the element before clicking.

        Example:

            >>> from stere.fields import Dropdown
            >>>
            >>> class CSSDropdown(Dropdown):
            >>>     def before(self):
            >>>         self.element.mouse_over()
        """
        pass

    def after(self) -> None:
        """Called automatically before methods with the `@use_after`
        decorator are called.

        Performer methods are decorated with @use_after.

        By default it does nothing. Override this method if an action must be
        taken after the method has been called.
        """
        pass

    def includes(self, value: str):
        """Will search every element found by the Field for a value property
        that matches the given value.
        If an element with a matching value is found, it's then returned.

        Useful for when you have non-unique elements and know a value is in
        one of the elements, but don't know which one.

        Arguments:
            value (str): A text string inside an element you want to find.

        Returns:
            element

        Example:

            >>> class PetStore(Page):
            >>>     def __init__(self):
            >>>         self.inventory = Link('xpath', '//li[@class="inv"]')
            >>>
            >>> pet_store = PetStore()
            >>> pet_store.inventory_list.includes("Kittens").click()

        """
        for item in self.element:
            if item.value == value:
                return item

    def value_contains(
        self, expected: str, wait_time: typing.Optional[int] = None,
    ) -> bool:
        """Check if the value of the Field contains an expected value.

        Arguments:
            expected (str): The expected value of the Field
            wait_time (int): The number of seconds to search.
                Default is Stere.retry_time.

        Returns:
            bool: True if the value was found, else False

        Example:

            >>> class PetStore(Page):
            >>>     def __init__(self):
            >>>         self.price = Link('xpath', '//li[@class="price"]')
            >>>
            >>> pet_store = PetStore()
            >>> assert pet_store.price.value_contains("19.19", wait_time=6)

        """
        return _retry(
            lambda: expected in self.value,
            retry_time=wait_time,
        )

    def value_equals(self, expected, wait_time: typing.Optional[int] = None):
        """Check if the value of the Field equals an expected value.

        Arguments:
            expected (str): The expected value of the Field
            wait_time (int): The number of seconds to search.
                Default is Stere.retry_time.

        Returns:
            bool: True if the value was found, else False

        Example:

            >>> class PetStore(Page):
            >>>     def __init__(self):
            >>>         self.price = Link('xpath', '//li[@class="price"]')
            >>>
            >>> pet_store = PetStore()
            >>> assert pet_store.price.value_equals("$19.19", wait_time=6)

        """
        return _retry(
            lambda: expected == self.value,
            retry_time=wait_time,
        )

    def find(self, wait_time: typing.Optional[int] = None):
        """Find the first matching element.

        Returns:
            Element

        Raises:
            ValueError - If more than one element is found.

        """
        found_elements = self.find_all(wait_time)
        if len(found_elements) >= 2:
            raise ValueError("Expected one element, found multiple")
        return found_elements[0]

    def find_all(self, wait_time: typing.Optional[int] = None):
        """Find all matching elements.

        Returns:
            list

        """
        return self._element.find(wait_time)



from ..decorators import stere_performer
from ..field import Field


@stere_performer('null_action', consumes_arg=False)
class Root(Field):
    """A simple wrapper over Field, it does not implement a performer method.
    Although Root has no specific behaviour, it can be useful when declaring a
    root for an Area or RepeatingArea.

    Example:

        >>> from stere.areas import RepeatingArea
        >>> from stere.fields import Root
        >>>
        >>>
        >>> collections = RepeatingArea(
        >>>     root=Root('xpath', '//table/tr'),
        >>>     quantity=Text('css', '.collection_qty'),
        >>> )
    """

    pass



from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('opposite', consumes_arg=False)
class Checkbox(Field):
    """Class with specific methods for handling checkboxes."""

    def __init__(self, *args, default_checked=False, **kwargs):
        super().__init__(*args, **kwargs)

        self.default_checked = default_checked

    def set_to(self, state: bool) -> None:
        """Set a checkbox to the desired state.

        Arguments:
            state (bool): True for check, False for uncheck

        Example:

            >>> confirm = Checkbox('id', 'selectme')
            >>> confirm.set_to(True)

        """
        if state:
            self.check()
        else:
            self.uncheck()

    def toggle(self) -> None:
        """If the checkbox is checked, uncheck it.
        If the checkbox is unchecked, check it.

        >>> confirm = Checkbox('id', 'selectme')
        >>> confirm.toggle()

        """
        if self.checked:
            self.uncheck()
        else:
            self.check()

    @use_after
    @use_before
    def check(self) -> None:
        """Use Splinter's check method."""
        self.find().check()

    @use_after
    @use_before
    def uncheck(self) -> None:
        """Use Splinter's uncheck method."""
        self.find().uncheck()

    def opposite(self) -> bool:
        """Switches the checkbox to the opposite of its default state.
        Uses the `default_checked` attribute to decide this.

        >>> confirm = Checkbox('id', 'selectme')
        >>> confirm.opposite()

        """
        if not self.default_checked:
            self.check()
        else:
            self.uncheck()
        return False



from typing import Optional

from splinter.driver.webdriver import WebDriverElement
from splinter.element_list import ElementList

from ..decorators import stere_performer
from ..field import Field


@stere_performer('null_action', consumes_arg=False)
class ShadowRoot(Field):
    """Field that finds the shadow-root of an element.

    Only useful as the root of an Area.

    Example:
        >>> address_form = Area(
        >>>     root=ShadowRoot('css', '#addressFormBlock'),
        >>>     address=Input('css', '#userAddress'),
        >>> )

    """

    def find_all(self, wait_time: Optional[int] = None) -> ElementList:
        """Get the shadowRoot element for any found elements.

        Returns:
            ElementList
        """
        found_elements = self._element.find(wait_time)

        shadow_roots = []
        for elem in found_elements:
            shadow_root = self.browser.execute_script(
                'return arguments[0].shadowRoot', elem._element,
            )
            shadow_roots.append(WebDriverElement(shadow_root, elem))

        return ElementList(
            shadow_roots, find_by=self.strategy, query=self.locator,
        )



from .clickable import Clickable
from ..decorators import stere_performer


@stere_performer('click', consumes_arg=False)
class Link(Clickable):
    """Convenience Class on top of Field.

    Implements `click()` as its performer.

    Example:

        >>> purchase = Link('id', 'buy_link')
        >>> purchase.click()

    """



import time

from .xhr_script import js_script
from ..browserenabled import BrowserEnabled


class XHRSpy(BrowserEnabled):
    """Spy for XMLHttpRequest instances in the browser.

    Allows scripts to block until network activity has stopped.

    Example:
        >>> from stere.browser_spy import XHRSpy
        >>>
        >>> spy = XHRSpy()
        >>> spy.add()
        >>> # Browser interaction
        >>> spy.wait_for_no_activity()
        >>> # More browser interaction

    Can also be used as a context manager, with add() called automatically:

    Example:
        >>> from stere.browser_spy import XHRSpy
        >>>
        >>> with XHRSpy() as spy:
        >>>     # Browser interaction
        >>>     spy.wait_for_no_activity()
        >>>     # More browser interaction

    """

    def __enter__(self):
        """As a context manager, the spy is added on enter."""
        self.add()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Nothing happens on exit."""
        pass

    def add(self):
        """Inject the spy onto the page.

        Tracks the active and total amount of requests.
        """
        self.browser.execute_script(js_script)

    def wait_for_no_activity(self, timeout: int = 30) -> bool:
        """Until timeout, keep checking if XHR is done.

        Requires the spy to already be present via FetchSpy.add()

        Arguments:
            timeout (int): Number of seconds to wait

        Raises:
            TimeoutError
        """
        end = time.time() + timeout
        no_active: bool = False

        while not no_active:
            no_active = self.active == 0

            # If no XHR events, wait and retry.
            # This confirms no new requests were made.
            if no_active:
                time.sleep(BrowserEnabled.xhr_spy_sleep_time)
                no_active = self.active == 0

            if time.time() > end:
                raise TimeoutError(f'XHR took longer than {timeout} seconds.')

        return no_active

    @property
    def active(self) -> int:
        """Get the number of active XHR requests."""
        return self.browser.execute_script("return document.activeXHRrequests")

    @property
    def total(self) -> int:
        """Get the number of total XHR requests."""
        return self.browser.execute_script("return document.totalXHRrequests")



from stere import Stere

from .strategy import strategies
from .strategy import strategy


__all__ = [
    'strategy',
    'strategies',
]

if Stere.library == 'appium':
    from .appium import FindByAccessibilityId
    from .appium import FindByAndroidUIAutomator
    from .appium import FindByIOSClassChain
    from .appium import FindByIOSUIPredicate
    from .appium import FindByIOSUIAutomation

    __all__ += [
        'FindByAccessibilityId',
        'FindByAndroidUIAutomator',
        'FindByIOSClassChain',
        'FindByIOSUIPredicate',
        'FindByIOSUIAutomation',
    ]

elif Stere.library == 'splinter':
    from .splinter import FindByCss
    from .splinter import FindById
    from .splinter import FindByName
    from .splinter import FindByTag
    from .splinter import FindByText
    from .splinter import FindByValue
    from .splinter import FindByXPath
    from .splinter import add_data_star_strategy

    __all__ += [
        'FindByCss',
        'FindByXPath',
        'FindByTag',
        'FindByName',
        'FindByText',
        'FindById',
        'FindByValue',
        'add_data_star_strategy',
    ]



from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('click', consumes_arg=False)
class Clickable(Field):
    """Convenience Class on top of Field.

    Implements `click()` as its performer.
    """

    @use_after
    @use_before
    def click(self) -> None:
        """Use Splinter's click method.

        Example:

            >>> purchase = Clickable('id', 'buy_button')
            >>> purchase.click()

        """
        self.is_visible()
        self.is_clickable()
        self.find().click()



from collections import UserList

from .area import Area
from ..utils import rgetattr


class Areas(UserList):
    """Searchable collection of Areas.

    Behaves like a list.
    """

    def append(self, item: Area) -> None:
        """Add a new Area to the container.

        Raises:
            TypeError: If a non-Area object is given.

        """
        if not isinstance(item, Area):
            raise TypeError(
                f"{item} is not an Area. "
                "Only Area objects can be inside Areas.",
            )

        self.data.append(item)

    def containing(self, field_name: str, field_value: str):
        """Search for Areas where the Field's value
        matches the expected value and then returns an Areas object with all
        matches.

        Arguments:
            field_name (str): The name of the Field object.
            field_value (str): The value of the Field object.

        Returns:
            Areas: A new Areas object with matching results

        Example:

            >>> class Inventory(Page):
            >>>     def __init__(self):
            >>>         self.items = RepeatingArea(
            >>>             root=Root('xpath', '//my_xpath_string'),
            >>>             description=Text('xpath', '//my_xpath_string')
            >>>         )
            >>>
            >>> def test_stuff():
            >>>     # Ensure 10 items have a price of $9.99
            >>>     inventory = Inventory()
            >>>     found_areas = inventory.items.areas.containing(
            >>>         "price", "$9.99")
            >>>     assert 10 == len(found_areas)

        """
        containing = Areas()
        for area in self:
            field = rgetattr(area, field_name)

            if field.value == field_value:
                containing.append(area)

        return containing

    def contain(self, field_name: str, field_value: str) -> bool:
        """Check if a Field in any Area contains a specific value.

        Arguments:
            field_name (str): The name of the Field object.
            field_value (str): The value of the Field object.

        Returns:
            bool: True if matching value found, else False

        Example:

            >>> class Inventory(Page):
            >>>     def __init__(self):
            >>>         self.items = RepeatingArea(
            >>>             root=Root('xpath', '//div[@id='inventory']'),
            >>>             description=Text('xpath', './td[1]')
            >>>         )
            >>>
            >>> def test_stuff():
            >>>     inventory = Inventory()
            >>>     assert inventory.items.areas.contain(
            >>>         "description", "Bananas")

        """
        for area in self:
            field = rgetattr(area, field_name)

            if field.value == field_value:
                return True

        return False



import typing

from .element_strategy import ElementStrategy
from .strategy import strategy


class AppiumBase(ElementStrategy):
    def _find_all(self, wait_time: typing.Optional[int] = None):
        """Find from inside a parent element."""
        parent = self.parent_locator or self.browser
        func = getattr(
            parent, f'find_elements_by_{self.strategy}')
        return func(self.locator)


@strategy('accessibility_id')
class FindByAccessibilityId(AppiumBase):
    strategy = 'accessibility_id'


@strategy('android_uiautomator')
class FindByAndroidUIAutomator(AppiumBase):
    strategy = 'android_uiautomator'


@strategy('ios_class_chain')
class FindByIOSClassChain(AppiumBase):
    strategy = 'ios_class_chain'


@strategy('ios_predicate')
class FindByIOSUIPredicate(AppiumBase):
    strategy = 'ios_predicate'


@strategy('ios_uiautomation')
class FindByIOSUIAutomation(AppiumBase):
    strategy = 'ios_uiautomation'



from stere import Stere

from .field import Field
from .generic.root import Root
from .generic.text import Text

__all__ = [
    'Field',
    'Root',
    'Text',
]

if Stere.library == 'appium':
    from .appium.button import Button
    from .appium.input import Input

    __all__ += [
        'Button',
        'Input',
    ]

elif Stere.library == 'splinter':
    from .splinter.button import Button
    from .splinter.checkbox import Checkbox
    from .splinter.dropdown import Dropdown
    from .splinter.input import Input
    from .splinter.link import Link
    from .splinter.money import Money
    from .splinter.shadow_root import ShadowRoot

    __all__ += [
        'Button',
        'Checkbox',
        'Dropdown',
        'Input',
        'Link',
        'Money',
        'ShadowRoot',
    ]



from stere.strategy import strategies


def build_element(desired_strategy: str, locator: str, parent_locator=None):
    """Get an element strategy instance."""
    known = strategies.keys()

    if desired_strategy in known:
        element_class = strategies[desired_strategy]
        return element_class(desired_strategy, locator, parent_locator)

    raise ValueError(
        f'The strategy "{desired_strategy}" is not in {list(known)}.')



import os

from setuptools import find_packages, setup


def read(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, 'r') as f:
        return f.read()


setup(
    name="stere",
    version="0.28.0",
    description="A nice way of implementing the Page Object pattern.",
    long_description=read('README.rst'),
    author="Joshua Fehler",
    author_email="jsfehler@gmail.com",
    license="MIT",
    url="https://github.com/jsfehler/stere",
    packages=find_packages(),
    install_requires=[
        'py-moneyed==1.0',
    ],
    extras_require={
        'splinter': ['splinter==0.14.0'],
    },
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ),
)



import typing
from functools import wraps


def stere_performer(
    method_name: str,
    consumes_arg: bool = False,
) -> typing.Callable:
    """Wrap a class to associate method_name with the perform() method.

    Associating a method with perform allows the class to be fully used
    by Area objects via Area.perform().

    Arguments:
        method_name (str): The name of the method to perform
        consumes_args (bool): True if the method takes an argument, else False

    In the following example, when ``Philosophers().diogenes_area.perform()``
    is called, ``DiogenesButton.philosophize()`` is called.

    Example:

        >>> @stere_performer('philosophize', consumes_arg=False)
        >>> class DiogenesButton(Field):
        >>>     def philosophize(self):
        >>>         print("As a matter of self-preservation, ")
        >>>         print("a man needs good friends or ardent enemies, ")
        >>>         print("for the former instruct him and the latter ")
        >>>         print("take him to task.")
        >>>
        >>>
        >>> class Philosophers(Page):
        >>>     def __init__(self):
        >>>         self.diogenes_area = Area(
        >>>             quote_button=DiogenesButton('id', 'idDio'),
        >>>             next_button=Button('id', 'idNext'),
        >>>         )
        >>>
        >>>
        >>> Philosophers().diogenes_area.perform()
    """
    def wrapper(cls):
        class Performer(cls):
            def perform(self, value=None):
                """Run the method designated as the performer"""
                performer = getattr(self, method_name)
                if consumes_arg:
                    performer(value)
                else:
                    performer()
                return self.returns

        # Preserve original class name and doc
        Performer.__name__ = cls.__name__
        Performer.__doc__ = cls.__doc__
        Performer.consumes_arg = consumes_arg
        return Performer
    return wrapper


def use_before(func: typing.Callable) -> typing.Callable:
    """When used on a method in a Field, the following will occur before
    the decorated method is called:
    - The Field's before() method will be called.
    - Any listeners registered to the 'before' event will be called.

    Example:

        >>> class TransformingButton(Field):
        >>>     def before(self):
        >>>         print('Autobots! Transform and...')
        >>>
        >>>     @use_before
        >>>     def roll_out(self):
        >>>         print('roll out!')
        >>>
        >>> tf = TransformingButton()
        >>> tf.roll_out()
        >>>
        >>> "Autobots! Transform and..."
        >>> "roll out!"
    """
    @wraps(func)
    def wrapper(obj, *inner_args, **inner_kwargs):
        obj.before()
        obj.emit('before')
        return func(obj, *inner_args, **inner_kwargs)
    return wrapper


def use_after(func: typing.Callable) -> typing.Callable:
    """When used on a method in a Field, the following will occur after
    the decorated method is called:
    - The Field's after() method will be called.
    - Any listeners registered to the 'after' event will be called.

    Example:

        >>> class TransformingButton(Field):
        >>>     def after(self):
        >>>         print('rise up!')
        >>>
        >>>     @use_after
        >>>     def transform_and(self):
        >>>         print('Decepticons, transform and...')
        >>>
        >>> tf = TransformingButton()
        >>> tf.transform_and()
        >>>
        >>> "Decepticons, transform and..."
        >>> "rise up!"
    """
    @wraps(func)
    def wrapper(obj, *inner_args, **inner_kwargs):
        result = func(obj, *inner_args, **inner_kwargs)
        obj.after()
        obj.emit('after')
        return result
    return wrapper



# Javascript intended to be opened as a string and injected into the browser.
js_script = """
// Number of fetch events that are active.
document.activeFetchEvents = 0;

// Number of fetch events that have been processed.
document.totalFetchEvents = 0;

// Store original fetch
var origFetch = window.fetch;

// Patch Fetch to record active requests
window.fetch = async (...args) => {

  // Increment counters before making call.
  document.activeFetchEvents += 1;
  document.totalFetchEvents += 1;

  // Fetch
  const response = await origFetch(...args);

  // Decrement counter on completion
  response
    .clone()
    .blob()
    .then(r => {document.activeFetchEvents -= 1})
    .catch(err => console.error(err))
  ;

  // return original response
  return response;

};
"""



from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('click', consumes_arg=False)
class Button(Field):
    """Convenience Class on top of Field.

    Implements `click()` as its performer.
    """

    @use_after
    @use_before
    def click(self):
        """Use Appium's click method.

        Example:

        >>> purchase = Button('id', 'buy_button')
        >>> purchase.click()

        """
        self.find().click()



from ..decorators import stere_performer
from ..field import Field


@stere_performer('null_action', consumes_arg=False)
class Text(Field):
    """A simple wrapper over Field, it does not implement a performer method.
    Although Text has no specific behaviour, it can be useful when declaring
    that a Field should just be static Text.

    Example:

        >>> from stere.fields import Text
        >>>
        >>>
        >>> self.price = Text('id', 'item_price')
    """

    pass



from .area import Area
from .areas import Areas
from .repeating import Repeating
from .repeating_area import RepeatingArea

__all__ = [
    'Area',
    'Areas',
    'Repeating',
    'RepeatingArea',
]



import platform
import typing

from selenium.webdriver.common.keys import Keys

from ..decorators import stere_performer, use_after, use_before
from ..field import Field


@stere_performer('fill', consumes_arg=True)
class Input(Field):
    """Convenience Class on top of Field.

    Uses Splinter's input method.

    Arguments:
        default_value (str): When Input.fill() is called with no arguments,
            this value will be used instead.
    """

    def __init__(self, *args, default_value=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.default_value = default_value

    @use_after
    @use_before
    def fill(self, value: typing.Optional[str] = None) -> None:
        """Use Splinter's fill method.

        Arguments:
            value (str): The text to enter into the input.

        Example:

            >>> first_name = Input('id', 'fillme')
            >>> first_name.fill('Joseph')

        """
        if value is None and self.default_value:
            value = self.default_value
        self.find().fill(value)

    def highlight(self):
        """Highlight the text content in an input element."""
        system = platform.system()

        if system == 'Darwin':  # OSX
            value = Keys.COMMAND + 'a'
        else:
            value = Keys.CONTROL + 'a'

        self.type(value)
